package s16;

import static s16.Queens.QueenBoardBasic;

import javafx.application.Application;
import javafx.scene.Node;
import javafx.scene.Scene;
import javafx.scene.image.Image;
import javafx.scene.layout.GridPane;
import javafx.scene.paint.ImagePattern;
import javafx.scene.shape.Rectangle;
import javafx.stage.Stage;

public class QueensGui extends Application {
  private static int size;
  private static int slowdownMs;

  public static final String TITLE = "Queens problem";
  public static final double MIN_HEIGHT = 300;
  public static final double MIN_WIDTH  = 300;
  public static final double INIT_HEIGHT = 600;
  public static final double INIT_WIDTH  = 600;

  private Stage primaryStage;
  private GridPane board;
  private QueenBoardForFX queenBoard;

  public static void main(String[] args) {
    size=6;
    slowdownMs=1500;
    if (args.length != 0)
      size = Integer.parseInt(args[0]);
    if (args.length > 1)
      slowdownMs = Integer.parseInt(args[1]);
    launch(args);
  }

  private void solve(int size) {
    queenBoard = new QueenBoardForFX(size);
    System.out.println("Start solving");
    if (Queens.isSolvable(queenBoard))
      System.out.println("Found !\n" + queenBoard);
    else
      System.out.println("Not found !");
  }

  @Override
  public void stop() {
    System.exit(0);
  }

  @Override
  public void start(Stage primaryStage) {

    this.primaryStage = primaryStage;
    this.primaryStage.setTitle(TITLE);

    initRootLayout();

    Thread thread = new Thread(() -> {
      // System.out.println("On showing event !");
      // Solve a problem
      solve(size);
    });

    primaryStage.setOnShown(windowEvent -> thread.start());

    primaryStage.show();
  }

  private void initRootLayout() {
    primaryStage.setMinWidth(MIN_WIDTH);
    primaryStage.setMinHeight(MIN_HEIGHT);

    initBoard(size);
    Scene scene = new Scene(board);
    primaryStage.setScene(scene);
    primaryStage.setWidth(INIT_WIDTH);
    primaryStage.setHeight(INIT_HEIGHT);
  }

  private void initBoard(int size) {
    board = new GridPane();
    for (int i = 0; i < size; i++) {
      for (int j = 0; j < size; j++) {
        QueensSquare square = ((i + j) % 2 == 0) ?
            new QueensSquare(SquareState.BRIGHT_EMPTY) :
              new QueensSquare(SquareState.DARK_EMPTY);
            board.add(square, i, j);
            square.widthProperty().bind(board.widthProperty().divide(size));
            square.heightProperty().bind(board.heightProperty().divide(size));
      }
    }
  }

  private void updateBoard() {
    for (Node node : board.getChildren()) {
      int col=GridPane.getColumnIndex(node), row=GridPane.getRowIndex(node);
      QueensSquare qnode = (QueensSquare) node;
      qnode.updateAgainst(row, col);
    }
  }

  //==============================================================
  private class QueenBoardForFX extends QueenBoardBasic {
    public QueenBoardForFX(int dim) {
      super(dim);
    }
    @Override public void putQueen(int row, int col) {
      mySleep(slowdownMs);
      super.putQueen(row, col);
      updateBoard();
    }
    @Override public void removeQueen(int row, int col) {
      mySleep(slowdownMs);
      super. removeQueen(row, col);
      updateBoard();
    }
    private void mySleep(int ms) {
      try {
        Thread.sleep(ms);
      } catch (InterruptedException e) {
        Thread.currentThread().interrupt();
      }
    }
    public boolean isDark(int row, int col) {
      return (row + col) % 2 != 0;
    }
  }
  //==============================================================
  public enum SquareState {
    DARK_TAKEN           ("/resources/dark-taken.jpg"),
    BRIGHT_TAKEN         ("/resources/bright-taken.jpg"),
    DARK_EMPTY           ("/resources/dark-empty.jpg"),
    BRIGHT_EMPTY         ("/resources/bright-empty.jpg"),
    DARK_ATTACKED_EMPTY  ("/resources/dark-empty-attacked-fb.jpg"),
    BRIGHT_ATTACKED_EMPTY("/resources/bright-empty-attacked-fb.jpg");

    private final String path;

    SquareState(String path) {
      this.path = path;
    }
  }
  //==============================================================
  public class QueensSquare extends Rectangle {

    public QueensSquare(SquareState state) {
      update(state);
    }

    public void updateAgainst(int row, int col) {
      if (queenBoard.isDark(row, col)) {
        if(queenBoard.isQueenThere(row, col))
          update(SquareState.DARK_TAKEN);
        else if (queenBoard.isSquareAttacked(row, col))
          update(SquareState.DARK_ATTACKED_EMPTY);
        else 
          update(SquareState.DARK_EMPTY);
      } else {
        if(queenBoard.isQueenThere(row, col))
          update(SquareState.BRIGHT_TAKEN);
        else if (queenBoard.isSquareAttacked(row, col))
          update(SquareState.BRIGHT_ATTACKED_EMPTY);
        else 
          update(SquareState.BRIGHT_EMPTY);
      }
    }

    private void update(SquareState state) {
      try {
        this.setFill(new ImagePattern(new Image(QueensSquare.class.getResourceAsStream(state.path))));
      } catch(RuntimeException e) {
        System.err.println("can't find image resources for QueensGui...");
        throw e;
      }
    }
  }

}
